/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Feb 14, 2011
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef TRAVERSATORS_CEDGE_COMM_VERTICES_NORMALS_HPP_
#define TRAVERSATORS_CEDGE_COMM_VERTICES_NORMALS_HPP_


#include "libsierpi/stacks/CFBStacks.hpp"

namespace sierpi
{
namespace travs
{


/**
 * adaptive refinement
 */
template <typename CKernelClass, typename CCellData, typename CEdgeData>
class CEdgeComm_VertexCoords_Normals
{
public:
	/**
	 * implementation of kernel class
	 */
	CKernelClass cKernelClass;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename CKernelClass::TVertexScalar TVertexScalar;
	typedef CEdgeComm_VertexCoords_Normals<CKernelClass, CCellData, CEdgeData> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStack<char> *structure_stack_out;
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * lifo and fifo iterator for element data
	 */
	CStackReaderTopDown<CCellData> element_data_lifo;
	CStackReaderBottomUp<CCellData> element_data_fifo;

	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<CEdgeData> *edge_data_comm_left_edge_stack;
	CStack<CEdgeData> *edge_data_comm_right_edge_stack;

	/**
	 * pointer to stack used for edge communication in forward order
	 */
	CStack<CEdgeData> *edge_comm_buffer;

#define SIERPI_PARALLEL_CODE	1
#include "auto/CTraversator_EdgeComm_VertexCoords_Normals_autogenerated.hpp"
#undef SIERPI_PARALLEL_CODE
#include "../CTraversatorClassInc_SetupTwoPass.hpp"


	void actionFirstPass(
			CFBStacks<char> &p_structure_stacks,
			CFBStacks<CCellData> &p_cell_data_stacks,
			CStack<CEdgeData> &p_edge_data_comm_left_edge_stack,
			CStack<CEdgeData> &p_edge_data_comm_right_edge_stack,
			CStack<CEdgeData> &p_edge_comm_buffer
	)
	{
		assert(p_cell_data_stacks.getNumberOfElementsOnStack() != 0);

		assert(p_structure_stacks.direction == p_cell_data_stacks.direction);
		assert(p_structure_stacks.direction == CFBStacks<char>::FORWARD);

		structure_stack_out = &p_structure_stacks.backward;
		structure_stack_out->clear();

		assert(structure_stack_out->isEmpty());

		structure_lifo_in.setup(p_structure_stacks.forward);

		edge_data_comm_left_edge_stack = &p_edge_data_comm_left_edge_stack;
		edge_data_comm_right_edge_stack = &p_edge_data_comm_right_edge_stack;

		edge_comm_buffer = &p_edge_comm_buffer;
		element_data_lifo.setup(p_cell_data_stacks.forward);

		cKernelClass.traversal_pre_hook();

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
		(this->*sfcRecursiveMethod_Forward)(
					cTriangleFactory.vertices[0][0], cTriangleFactory.vertices[0][1],
					cTriangleFactory.vertices[1][0], cTriangleFactory.vertices[1][1],
					cTriangleFactory.vertices[2][0], cTriangleFactory.vertices[2][1],
					cTriangleFactory.hypNormal
		);
#else
		(this->*sfcRecursiveMethod_Forward)(
					cTriangleFactory.vertices[0][0], cTriangleFactory.vertices[0][1],
					cTriangleFactory.vertices[1][0], cTriangleFactory.vertices[1][1],
					cTriangleFactory.vertices[2][0], cTriangleFactory.vertices[2][1]
		);
#endif

		assert(structure_lifo_in.isEmpty());
		assert(element_data_lifo.isEmpty());
	}


	void actionSecondPass(
			CFBStacks<char> &p_structure_stacks,
			CFBStacks<CCellData> &p_cell_data_stacks,
			CStack<CEdgeData> &p_edge_data_comm_left_edge_stack,
			CStack<CEdgeData> &p_edge_data_comm_right_edge_stack,
			CStack<CEdgeData> &p_edge_comm_buffer
	)
	{
		assert(p_cell_data_stacks.getNumberOfElementsOnStack() != 0);

		edge_data_comm_left_edge_stack = &p_edge_data_comm_right_edge_stack;
		edge_data_comm_right_edge_stack = &p_edge_data_comm_left_edge_stack;

		edge_comm_buffer = &p_edge_comm_buffer;

		structure_lifo_in.setup(&p_structure_stacks.backward);
		element_data_fifo.setup(p_cell_data_stacks.forward);

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
		(this->*sfcRecursiveMethod_Backward)(
					cTriangleFactory.vertices[0][0], cTriangleFactory.vertices[0][1],
					cTriangleFactory.vertices[1][0], cTriangleFactory.vertices[1][1],
					cTriangleFactory.vertices[2][0], cTriangleFactory.vertices[2][1],
					cTriangleFactory.hypNormal
		);
#else
		(this->*sfcRecursiveMethod_Backward)(
					cTriangleFactory.vertices[0][0], cTriangleFactory.vertices[0][1],
					cTriangleFactory.vertices[1][0], cTriangleFactory.vertices[1][1],
					cTriangleFactory.vertices[2][0], cTriangleFactory.vertices[2][1]
		);
#endif
		
		cKernelClass.traversal_post_hook();

		assert(edge_comm_buffer->isEmpty());
		assert(edge_data_comm_left_edge_stack->isEmpty());
		assert(edge_data_comm_right_edge_stack->isEmpty());
		assert(structure_lifo_in.isEmpty());
		assert(element_data_fifo.isEmpty());

		structure_stack_out->clear();
	}
};

}
}

#endif
