/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Feb 14, 2011
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef TRAVERSATORS_CFLUX_COMM_DEPTH_RETURN_ADAPVITIVY_REQUESTS_HPP_
#define TRAVERSATORS_CFLUX_COMM_DEPTH_RETURN_ADAPVITIVY_REQUESTS_HPP_


#include "libsierpi/stacks/CSimulationStacks.hpp"
#include "libsierpi/stacks/CStackWriterTopDown.hpp"
#include "libsierpi/triangle/CTriangle_Factory.hpp"
#include "libsierpi/triangle/CTriangle_Normals.hpp"


namespace sierpi
{
namespace travs
{

/**
 * adaptive refinement
 */
template <
	class t_CKernelClass,
	typename t_CTypes
>
class CFluxComm_Depth_ReturnAdaptivityRequests
{
public:
	enum
	{
		ADAPTIVITY_NOTHING = 0,
		ADAPTIVITY_REFINE = 68,		// local refine request + forward information along hypotenuse
		ADAPTIVITY_COARSEN = 50	// local coarsening request + forward information along left/right edge
	};

	typedef t_CKernelClass CKernelClass;

	/**
	 * implementation of kernel class
	 */
	CKernelClass cKernelClass;

	typedef typename CKernelClass::TReduceValue TReduceValue;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename t_CTypes::CSimulationTypes::CCellData CCellData;
	typedef typename t_CTypes::CSimulationTypes::CEdgeData CEdgeData;
	typedef typename t_CTypes::CSimulationTypes::T T;
	typedef typename t_CTypes::CSimulationStacks CSimulationStacks;

	typedef CFluxComm_Depth_ReturnAdaptivityRequests<t_CKernelClass,t_CTypes> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStack<char> *structure_stack_out;
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * lifo and fifo iterator for element data
	 */
	CStackReaderTopDown<CCellData> element_data_lifo;
	CStackReaderBottomUp<CCellData> element_data_fifo;

	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<CEdgeData> *edge_data_comm_left_edge_stack;
	CStack<CEdgeData> *edge_data_comm_right_edge_stack;

	/**
	 * pointer to stack used for edge communication in forward order
	 */
	CStack<CEdgeData> *edge_comm_buffer;

	/**
	 * stack to store element wise refine/coarsen automata states
	 */
	CStackWriterTopDown<unsigned char> adaptivity_output_stream;

	/**
	 * repeat traversal to get conforming grid?
	 */
	bool repeat_traversal;

	/**
	 * depth limiter: min/max
	 */
	int depth_limiter_min;
	int depth_limiter_max;


	void setup_RootTraversator(
			int p_depth_limiter_min,
			int p_depth_limiter_max
	)
	{
		depth_limiter_min = p_depth_limiter_min;
		depth_limiter_max = p_depth_limiter_max;
	}


	CFluxComm_Depth_ReturnAdaptivityRequests()
#if DEBUG
		:
		depth_limiter_min(-1),
		depth_limiter_max(-1)
#endif
	{
	}


#if CONFIG_ENABLE_MPI_CLUSTER_MIGRATION

	void migration_send(int i_dst_rank)
	{
		cKernelClass.migration_send(i_dst_rank);
	}

	void migration_send_postprocessing(int i_dst_rank)
	{
		cKernelClass.migration_send_postprocessing(i_dst_rank);
	}

	void migration_recv(int i_src_rank)
	{
		cKernelClass.migration_recv(i_src_rank);
	}

#endif


#define SIERPI_PARALLEL_CODE	1
#include "auto/CTraversator_FluxComm_Depth_RetAdapt_autogenerated.hpp"
#undef SIERPI_PARALLEL_CODE



	/**
	 * recursive method for forward traversal
	 */
	TRecursiveMethod sfcRecursiveMethod_Forward;


	/**
	 * recursive method for backward traversal
	 */
	TRecursiveMethod sfcRecursiveMethod_Backward;

	/**
	 * the triangle factory for this traversal
	 */
	CTriangle_Factory cTriangleFactory;



	void action_FirstPass(
			CSimulationStacks *io_cSimulationStacks
	)
	{
		assert(io_cSimulationStacks->structure_stacks.direction == io_cSimulationStacks->cell_data_stacks.direction);
		assert(io_cSimulationStacks->structure_stacks.direction == CFBStacks<char>::FORWARD);

		structure_stack_out = &io_cSimulationStacks->structure_stacks.backward;
		structure_stack_out->clear();
		assert(structure_stack_out->isEmpty());

		structure_lifo_in.setup(io_cSimulationStacks->structure_stacks.forward);

		assert(!io_cSimulationStacks->structure_stacks.forward.isEmpty());

		edge_data_comm_left_edge_stack = &io_cSimulationStacks->edge_data_comm_edge_stacks.left;
		edge_data_comm_right_edge_stack = &io_cSimulationStacks->edge_data_comm_edge_stacks.right;

		edge_comm_buffer = &io_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(io_cSimulationStacks->cell_data_stacks.forward);

		assert(depth_limiter_min >= 0);
		assert(depth_limiter_max >= 0);

#if CONFIG_SUB_SIMULATION_EULER_MULTILAYER
		cKernelClass.traversal_pre_hook(io_cSimulationStacks->cell_data_stacks.getMaxNumberOfElements());
#else
		cKernelClass.traversal_pre_hook();
#endif

		(this->*sfcRecursiveMethod_Forward)(cTriangleFactory.recursionDepthFirstRecMethod);

		cKernelClass.op_edge_edge_stack(&(io_cSimulationStacks->edge_comm_buffer));

		assert(structure_lifo_in.isEmpty());
		assert(element_data_lifo.isEmpty());
	}


	void actionMiddlePass_computeClusterBorderCFL(
			CSimulationStacks *p_cSimulationStacks,
			TReduceValue *o_reduceValue
	)	{
		cKernelClass.updateTimestepSizeWithNetUpdates(p_cSimulationStacks->edge_data_comm_exchange_edge_stacks.right);
		cKernelClass.updateTimestepSizeWithNetUpdates(p_cSimulationStacks->edge_data_comm_exchange_edge_stacks.left);

		*o_reduceValue = cKernelClass.getCFL1TimestepSize();
	}


	/**
	 * second pass for parallel processing (exchange stacks)
	 */
	void action_SecondPass_Parallel(
		CSimulationStacks *io_cSimulationStacks,
		T i_timestep_size
	)
	{
		assert(!io_cSimulationStacks->structure_stacks.backward.isEmpty());

		structure_lifo_in.setup(io_cSimulationStacks->structure_stacks.backward);
		element_data_fifo.setup(io_cSimulationStacks->cell_data_stacks.forward);



		// clear "old" edge comm stacks
		io_cSimulationStacks->edge_data_comm_edge_stacks.clear();

		edge_data_comm_left_edge_stack = &io_cSimulationStacks->edge_data_comm_exchange_edge_stacks.right;
		edge_data_comm_right_edge_stack = &io_cSimulationStacks->edge_data_comm_exchange_edge_stacks.left;

		// backup previous edge comm elements to setup dummy edge comm data for adaptivity traversals
		io_cSimulationStacks->previous_edge_comm_elements_left = edge_data_comm_right_edge_stack->getNumberOfElementsOnStack();
		io_cSimulationStacks->previous_edge_comm_elements_right = edge_data_comm_left_edge_stack->getNumberOfElementsOnStack();

		edge_comm_buffer = &io_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(io_cSimulationStacks->cell_data_stacks.forward);

		// set the timestep size
		cKernelClass.setTimestepSize(i_timestep_size);

		size_t elements_on_stack = io_cSimulationStacks->cell_data_stacks.forward.getNumberOfElementsOnStack();
		io_cSimulationStacks->adaptivity_state_flag_stacks.backward.setStackElementCounter(elements_on_stack);

		adaptivity_output_stream.setup(
				&(io_cSimulationStacks->adaptivity_state_flag_stacks.backward),
				elements_on_stack
			);

		repeat_traversal = false;

		// SFC traversals
		(this->*sfcRecursiveMethod_Backward)(cTriangleFactory.recursionDepthFirstRecMethod);

		io_cSimulationStacks->valid_backward_adaptivity_state_flags = true;

		assert(io_cSimulationStacks->structure_stacks.backward.structure_isValidTriangle());

		io_cSimulationStacks->repeat_adaptivity_traversal_for_conforming_grid = repeat_traversal;

		// post hook
		cKernelClass.traversal_post_hook();

		assert(adaptivity_output_stream.isAtBottom());

		assert(edge_comm_buffer->isEmpty());
		assert(edge_data_comm_left_edge_stack->isEmpty());
		assert(edge_data_comm_right_edge_stack->isEmpty());
		assert(structure_lifo_in.isEmpty());
		assert(element_data_fifo.isEmpty());
	}



	/**
	 * second pass for serial traversal
	 */
	void action_SecondPass_Serial(
		CSimulationStacks *io_cSimulationStacks,
		T i_timestep_size
	)
	{
		structure_lifo_in.setup(io_cSimulationStacks->structure_stacks.backward);
		element_data_fifo.setup(io_cSimulationStacks->cell_data_stacks.forward);

		assert(!io_cSimulationStacks->structure_stacks.backward.isEmpty());

		edge_data_comm_left_edge_stack = &io_cSimulationStacks->edge_data_comm_edge_stacks.right;
		edge_data_comm_right_edge_stack = &io_cSimulationStacks->edge_data_comm_edge_stacks.left;

		// backup previous edge comm elements to setup dummy edge comm data for adaptivity traversals
		io_cSimulationStacks->previous_edge_comm_elements_left = edge_data_comm_right_edge_stack->getNumberOfElementsOnStack();
		io_cSimulationStacks->previous_edge_comm_elements_right = edge_data_comm_left_edge_stack->getNumberOfElementsOnStack();

		edge_comm_buffer = &io_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(io_cSimulationStacks->cell_data_stacks.forward);

		// set the timestep size
		cKernelClass.setTimestepSize(i_timestep_size);

		size_t elements_on_stack = io_cSimulationStacks->cell_data_stacks.forward.getNumberOfElementsOnStack();
		io_cSimulationStacks->adaptivity_state_flag_stacks.backward.setStackElementCounter(elements_on_stack);
		adaptivity_output_stream.setup(
				&(io_cSimulationStacks->adaptivity_state_flag_stacks.backward),
				elements_on_stack
			);

		repeat_traversal = false;

		// SFC traversals
		(this->*sfcRecursiveMethod_Backward)(cTriangleFactory.recursionDepthFirstRecMethod);

		io_cSimulationStacks->valid_backward_adaptivity_state_flags = true;
		io_cSimulationStacks->repeat_adaptivity_traversal_for_conforming_grid = repeat_traversal;

		// post hook
		cKernelClass.traversal_post_hook();

		assert(adaptivity_output_stream.isAtBottom());

		assert(edge_comm_buffer->isEmpty());
		assert(edge_data_comm_left_edge_stack->isEmpty());
		assert(edge_data_comm_right_edge_stack->isEmpty());
		assert(structure_lifo_in.isEmpty());
		assert(element_data_fifo.isEmpty());
	}



	/**
	 * setup traversator parameters (min/max depth)
	 */
	inline void setup_TraversatorParameters(
			int i_depth_limiter_min,
			int i_depth_limiter_max
	)
	{
		depth_limiter_min = i_depth_limiter_min;
		depth_limiter_max = i_depth_limiter_max;
	}



	/**
	 * setup the initial cluster traversal for the given factory
	 */
	inline void setup_Cluster(
			CFluxComm_Depth_ReturnAdaptivityRequests<t_CKernelClass, t_CTypes> &i_parent,
			CTriangle_Factory &i_cTriangleFactory
	)
	{
		// make sure that this is really a root node
		assert(i_cTriangleFactory.clusterTreeNodeType != CTriangle_Enums::NODE_ROOT_TRIANGLE);

		setup_sfcMethods(i_cTriangleFactory);

		sfcRecursiveMethod_Forward = getSFCMethod<SFC_METHOD_OLD_TO_NEW_EDGE_TYPES>(i_cTriangleFactory);
		sfcRecursiveMethod_Backward = getSFCMethod<(SFC_METHOD_NEW_TO_OLD_EDGE_TYPES | SFC_METHOD_FORCE_BACKWARD)>(i_cTriangleFactory);


		setup_TraversatorParameters(i_cTriangleFactory.minDepth, i_cTriangleFactory.maxDepth);

		/**
		 * SETUP KERNEL CLASS
		 */
		cKernelClass.setup_WithKernel(i_parent.cKernelClass);
	}



	/**
	 * setup the initial cluster traversal for the given factory
	 */
	inline void setup_sfcMethods(
			CTriangle_Factory &i_cTriangleFactory
	)
	{
		cTriangleFactory = i_cTriangleFactory;

		/*
		 * search for recursive function id for first pass
		 *
		 * all edges of type OLD  of this triangle have to be set to NEW
		 */
		sfcRecursiveMethod_Forward = getSFCMethod<SFC_METHOD_OLD_TO_NEW_EDGE_TYPES>(cTriangleFactory);

		/*
		 * search for recursive function id for second pass
		 *
		 * all edges of type NEW  of this triangle have to be set to OLD
		 */

		sfcRecursiveMethod_Backward = getSFCMethod<(SFC_METHOD_NEW_TO_OLD_EDGE_TYPES | SFC_METHOD_FORCE_BACKWARD)>(cTriangleFactory);

		setup_TraversatorParameters(i_cTriangleFactory.minDepth, i_cTriangleFactory.maxDepth);
	}

};

}
}

#endif /* CSTRUCTURE_ADAPTIVEREFINEMENT_HPP_ */
