/*
 * Copyright (C) 2012 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: May 12, 2012
 *      Author: Aditya ghantasala <shine.aditya@gmail.com>
 */

#ifndef VARIABLE_HPP_
#define VARIABLE_HPP_

#include <cmath>
#include <cassert>

#include <iostream>
#include <string.h>
#include <netcdfcpp.h>

#include "FindNorm.hpp"
#include "ReadData.hpp"
#include "CheckData.hpp"
#include "ReadNetcdfFile.hpp"



namespace benchMark {
template <typename T>
class CVerify;

}


template <typename T>
class benchMark::CVerify{

private:

	//Stores the norm every time the bench mark is performed
	T* norm;
	T* benchMarkTimeVector;


	std::string varName;

	// Variables used to store the quantities
	// from previous time a benchmark is performed


	int *counter;

	// Variable of the class which reads the NetCDF file
	benchMark::CReadNetcdfFile<T> ncFile;

	//Pointers to store the values read from the NetCDF file
	T* timeVector;
	T* xVector;
	T* yVector;
	T* heightMatrix;

	benchMark::CReadData<T> readData;



public:

	/**
	 * Empty Constructor for the Class Variable

	Verify(){

	}*/

	/**
	 * Method for setting up various parameters for the object of class Variable
	 *
	 * @param i_fileName			- Name of the NetCDF file containing the Benchmark variable
	 * @param i_timeVarableName		- Name of the Time variable/Dimension given in the NetCDF file.
	 * @param i_xVariableName		- Name of the X variable/Dimension given in the NetCDF file.
	 * @param i_yVariableName		- Name of the Y variable/Dimension given in the NetCDF file.
	 * @param i_heightVariableName	- Name of the Height variable/Dimension given in the NetCDF file.
	 */
	void setup(
			std::string i_fileName,
			std::string i_timeVarableName,
			std::string i_xVariableName,
			std::string i_yVariableName,
			std::string i_heightVariableName
	){
		varName = i_heightVariableName;


		counter = 0;

		//Initializing the positions

		std::cout << "Loading NetCDF data " << i_fileName << std::endl;

		//Reading the NetCDF File given
		ncFile.setup(i_fileName,i_timeVarableName,i_xVariableName,i_yVariableName,i_heightVariableName); //This also reads the file parameters into memory

		//Allocating memory to the Vector Pointers
		timeVector = new T[ncFile.getTimeVectorLength()];
		xVector = new T[ncFile.getXVectorLength()];
		yVector = new  T[ncFile.getYVectorLength()];
		heightMatrix = new T[ncFile.getTimeVectorLength()*ncFile.getXVectorLength()*ncFile.getYVectorLength()];

		ncFile.readFile(timeVector, xVector, yVector, heightMatrix);

		readData.setup(ncFile, timeVector, xVector, yVector, heightMatrix);
		//############ Allocating Memory to Norm and benchMarkTimeVector #############

		benchMarkTimeVector = new T[ncFile.getTimeVectorLength()];
		counter = new int[ncFile.getTimeVectorLength()];

		//Allocating the memory for norm vector
		norm = new T[ncFile.getTimeVectorLength()];

		//Making all the norm variables zero
		for(int i=0;i<ncFile.getTimeVectorLength();i++){
			norm[i] = 0;
		}
		//Making the counter Zero
		for(int i=0;i<ncFile.getTimeVectorLength();i++){
			counter[i] = 1;
		}

	}

	/**
	 * Method to read the Benchmark Value from the NetCDF File
	 *
	 * @param i_simulationTime	- Global Simulation time of the Simulation.
	 * @param i_xCordinate		- X Coordinate where a benchmark should be carried out.
	 * @param i_yCordinate		- Y Coordinate where a benchmark should be carried out.
	 * @param o_benchMarkValue	- Pointer to a variable where the read benchmark Variable should be stored.
	 */
	void getBenchMarkNodalData(
			T i_simulationTime,
			T i_xCordinate,
			T i_yCordinate,
			T i_level_of_detail,
			T* o_benchmark_value
	)	{
		if (i_simulationTime > timeVector[0] && i_simulationTime<5+timeVector[ncFile.getTimeVectorLength()-1])
		{
			if(ncFile.getYVectorLength() == 1)
				i_yCordinate = 0;

			if (!readData.getBenchMarkData(
					i_simulationTime,
					i_xCordinate,
					i_yCordinate,
					o_benchmark_value
			))
			{
					*o_benchmark_value = 0;
			}
		}
		else {
			*o_benchmark_value = 0;
		}

	}

	/**
	 * Method to complete the Benchmark and output various norms and other parameters
	 * Calculated as a part of the Benchmark.
	 */
	void completeBenchmark(){

		int i;

		//		tools::Logger logger(0, "BENCHMARKING RESULTS");
		int l_length = ncFile.getTimeVectorLength();

		//Checking if the all counter elements are more than zero
		for(i=0;i<l_length;i++){
			assert(counter[i]>0);
		}

		//		logger.printNorms(norm,timeVector,l_length,counter,"Profiles");

	}


	//Destructor
	~CVerify(){

	}


};



#endif /* VARIABLE_HPP_ */
